/*  ------------------------------------------------ */
/*  grid correction study and determination          */
/*  jfberar   berar@esrf.fr                          */
static char *cvs_id =" $Id: bm2grid.c,v 1.3 2002/07/22 12:31:40 berar Exp $ ";

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
/*-----------------------------------------------------------*/
#include "prg_def.h"

#include "xy_def.h"

#include "wv_def.h"
#include "gel_def.h"
#include "edf_def.h"
#include "dmi_def.h"
int cr_mode=0;
/*-----------------------------------------------------------*/
#ifndef _INCLUDE_HPUX_SOURCE
#include <time.h> 
#endif
/*-----------------------------------------------------------*/
#define TITLE "detector grid analyse"
#define RELEASE "berar@esrf.fr (d2am), "


/* default values */
typedef struct
{   IMAGE * img;          
    float * weight;
    float scale; 
	 float added_raw_scale;           
	 int   shift_x, shift_y;
	 int   added_raw;
	 int   valid;
	 } RESULT;


typedef struct
{   short x,y;
    unsigned z;
    } GRID_PIXEL;
    
typedef struct
{   GRID_PIXEL *pix;
    int count;
    int alloc;
    double x,y,z;
    double cx,cy;
    double d;
    int  nx, ny;
    char ok;
    } GRID_SUM;
/*-----------------------------------------------------------*/
/* declarations globales */
IMAGE image;
char     *prg_name, *grid_name;
char     *r_def = NULL;
IMAGE    *raw_img = NULL; 
GRID_SUM *grid_data; 
int     grid_data_count=0;
float   grid_level = 10;
float   grid_error = 0.25;
int     grid_count = 10;
float   grid_size = 1;
double  gridorig_x, gridorig_y;
double  gridrot[4];
#define gridrot_xx gridrot[0]
#define gridrot_xy gridrot[1]
#define gridrot_yx gridrot[2]
#define gridrot_yy gridrot[3]
double  gridnorm;
int world_x, world_y;

extern GRID_COR grid;   /* only for memory, done also in xy_def.h */
#define DEG 5
#define COR (((DEG+2)*(DEG+1))/2)
int     xcor_orig, ycor_orig;
T_DATA t_mode = INVALID_DATA;
int    set_limit = 0;
int  low_limit  = 0x80000000;
int  high_limit = 0x80000000; 
int    dformat     = 0;
int    print       = 2;
FILE *xmgr;
#if defined ( __STDC__ ) || defined ( MAC_OS )
static time_t secs_now;
#else
static long   secs_now;
#endif
char  *str_time;
/*-----------------------------------------------------------*/
FILE *inf;
/*-----------------------------------------------------------*/
void  usage( char *name);
void  parameter( FILE * f_ini);
void  put_image( char *name, char *path, IMAGE *img);
/*-----------------------------------------------------------*/
void read_grid_image   (char * buffer);
void read_grid   (char * buffer);


GRID_SUM * get_grid_data(void);
void  find_grid_vectors (void);
int   find_grid_nodes   (double tol);
int  clean_grid_nodes   (double err);
void   cor_grid_spots (int size);
int    ref_grid_spots (int deg, int doprint);
double err_grid_spots (int doprint);
void   img_grid_spots (char *name);
double ini_grid_cor   (void);
double opt_grid_cor   (int found, double err);
void   prt_grid_cor   (double *cor, double *new, char *name, int size, int offset, int doprint);
int chol_part (double * mat, double * vect, int n);

int explore(GRID_SUM *sum,  int istart, int jstart, int level, int *tot_pix, int *inside);
/*-----------------------------------------------------------*/
void grid_image(char * buffer);
void grid_print(char * buffer);
void grid_input(char * buffer);
void grid_check(char * buffer);
/*-----------------------------------------------------------*/
void usage( char *name)
{
  fprintf(stderr, "usage : %s data_file\n\
   if no data_file is given, the program try to use bm2grid.ini\n\
   else it asks for the data file\n\
   This file contains the following input keys\n\
   @GRID_IMAGE=image_file_name   : calculate grid correction\n\
   @GRID_PRINT=output_file_name  : print grid correction parameter\n\
   @GRID_INPUT=input_file_name   : read data from previous file\n\
   @GRID_CHECK=image_file_name   : correct image and check new grid\n\
   @GRID_INPUT=input_file_name    : apply grid correction\n\
   img_files contain only one frame (dark and cosmic corrected)\n\
   see bm2grid.html for parameters\n\
   it use the following dedicated tags in bm2grid.ini\n\
   GRID_LEVEL=10   trigger value used for detecting spots\n\
   GRID_ERROR=0.25 tolerance on spot position in translation definition\n\
   GRID_COUNT=10   minimum pixels number  in valid spots\n\
   GRID_SIZE=1.0   step size of the grid (to calculate pixel size)\n\
   Inside the program the image are kept as integers with 32 bits\n\
   see also parameters in %s.ini file\n", prg_name, prg_name );
   exit(3);
   }
/*-----------------------------------------------------------*/
void parameter( FILE *f_ini)
{   char buffer[100], c , *d, *e;
    char *tag, *arg;
    while (!feof(f_ini))
    {   c = fgetc( f_ini );
	     ungetc(c, f_ini);
		  if (c  == '@') return;
	     fgets(buffer, sizeof(buffer)-1, f_ini);
        if (feof(f_ini)) break;
        d = buffer;
        while (*d == ' ' || *d == '\n' || *d == '\r') d++;
        if (*d=='#' || *d==0) continue;
        if ( prg_initag(d, &tag, &arg) == 0 ) continue;
        if ( xy_inipar(&image, tag, arg)) continue;
        if ( prg_iniint(tag, arg, "LOG", &log_mode)) continue;
        if ( prg_inistr(tag, arg, "DEF", &r_def)) continue;
        if ( prg_iniint(tag, arg, "PRINT", &print)) continue;
        if ( prg_iniint(tag, arg, "LOW_LIMIT", &low_limit)) continue;
        if ( prg_iniint(tag, arg, "HIGH_LIMIT", &high_limit)) continue;

        if ( prg_iniint(tag, arg, "GRID_COUNT", &grid_count)) continue;
        if ( prg_iniflt(tag, arg, "GRID_LEVEL", &grid_level)) continue;
        if ( prg_iniflt(tag, arg, "GRID_ERROR", &grid_error)) continue;
        if ( prg_iniflt(tag, arg, "GRID_SIZE", &grid_size)) continue;

        if ( grid_ini( &grid_cor, tag, arg)) continue;

        if ( (e=strstr(d, "T_DATA")) != 0 )
        {   e+=6;
            while (*e==' ' || *e=='=') e++;
            if (strstr(e, "T_INT32") )  t_mode = T_INT32;
  	    else if (strstr(e, "T_FLOAT") )  t_mode = T_FLOAT;
  	    else 
  	    {   prg_printf(" invalid %s mode in T_DATA, reset to INVALID_DATA\n", d);
  	        t_mode = INVALID_DATA; d = "INVALID_DATA";
  	        }
  	    prg_printf(" T_DATA= %s\n", d);
            }
       
       if (strcmp(tag, "P_EXCLUDED") == 0 )  { excl_ini( P_EXCL, arg ); continue; }
       if (strcmp(tag, "C_EXCLUDED") == 0 )  { excl_ini( C_EXCL, arg ); continue; }
       if (strcmp(tag, "R_EXCLUDED") == 0 )  { excl_ini( R_EXCL, arg ); continue; }
       if (strcmp(tag, "T_EXCLUDED") == 0 )  { excl_ini( T_EXCL, arg ); continue; }
       }
   }
/*-----------------------------------------------------------*/
int main ( int argc, char ** argv )
{   char name[200], buffer[100], *d;
    enum { USE_ARGS, ASK_NAME, USE_INI } mode;
    long i;

#ifdef __POWERPC__
    prg_name = "bm2grid";
    argc = 0;
#else
    prg_name = argv[0];
    argc--; argv++;
#endif    
    xy_chktype();
    r_def = ".spe";
	 if (strstr(prg_name, "spe") == prg_name) r_def = ".spe";
    if (strstr(prg_name, "edf") == prg_name) r_def = ".edf";
    if (strstr(prg_name, "gel") == prg_name) r_def = ".gel";
    memset(&image, 0, sizeof(IMAGE));
    prg_printf("## %s : %s\n## %s %s\n", prg_name, TITLE, RELEASE, __DATE__);
    mode = (argc == 0) ?  ASK_NAME : USE_ARGS;
    if ( argc == 1 ) print |= 16;
    inf = NULL;
    if (mode == ASK_NAME)
    {   strcpy(name,prg_name);
        strcat(name,".ini");
        inf=fopen(name,"r");
        if (inf) mode= USE_INI;
        }
    switch (mode)
    {  case ASK_NAME :
            fprintf(stderr, "\n data file name : ");
	         buffer[0]=0;
	         fgets(buffer, sizeof(buffer)-1, stdin);
/*	         if ( redir ) puts(buffer); */
              name[0]=0;
              sscanf(buffer," %s", name);
              break;
       case USE_ARGS : strcpy(name, *argv);
             argc--; argv++;
	   }
   if ( name[0] == '-' ) usage("");
	if (!inf) inf = fopen(name,"r");
	if (!inf) v_error(__FILE__, "unable to open '%s' data file!", name);
	parameter(inf);
	while ( !feof(inf) )
	{  buffer[0]=0;
	   fgets(buffer, 200, inf);
      if ((d=strrchr(buffer, '\n'))) *d=0;
	   if (buffer[0]==0 || buffer[0] == '#') continue;
		else if (strstr(buffer, "@GRID_IMAGE") == buffer)  grid_image(buffer);  
		else if (strstr(buffer, "@GRID_PRINT") == buffer)  grid_print(buffer);
		else if (strstr(buffer, "@GRID_INPUT") == buffer)  grid_input(buffer);
		else if (strstr(buffer, "@GRID_CHECK") == buffer)  grid_check(buffer);
		else v_error(__FILE__,"unable to use statement '%s'", buffer);
		}
   PRG_EXIT
   return 0;
   }
/*-----------------------------------------------------------*/
double ini_grid_cor(void)
{  int j;
   double err;
   grid_struct_def(&grid_cor,COR,xcor_orig,ycor_orig);
   cor_grid_spots(3);
   err=err_grid_spots(1);
   return err;
   }
/*-----------------------------------------------------------*/
static char * coef_grid_name[] = 
   {"1","x","y","xx","xy","yy","xxx","xxy","xyy","yyy",
             "xxxx","xxxy","xxyy","xyyy","yyyy",
             "xxxxx","xxxxy","xxxyy","xxyyy","xyyyy","yyyyy"};
/*-----------------------------------------------------------*/
void prt_grid_cor(double *cor, double *new, char *name, int size, int offset,int doprint  )
{  int i;
   if (doprint) printf("prt_grid_cor axe %s (%d)\n", name, size);
   for (i=0;i<size;i++)
   {  if (doprint) printf(" %5s : %10.4lg %s", coef_grid_name[i], 
                new[i+offset], i%4==3?"\n":"   ");
      cor[i] = new[i+offset];
      }
   if (doprint) printf("\n");
   }
/*-----------------------------------------------------------*/
/*  the following routines are used by "@GRID_IMAGE"         */
/*-----------------------------------------------------------*/
void grid_image (char * buffer)
{   int found, fclean;
    double err;
    read_grid_image (buffer);
    world_x=raw_img->x_size;
    xcor_orig=world_x/2;
    world_y=raw_img->y_size;
    ycor_orig=world_y/2;
    grid_data = get_grid_data();  
    find_grid_vectors();
    found = find_grid_nodes(grid_error);
    err=ini_grid_cor();
    err=opt_grid_cor(found,err);
    fclean = clean_grid_nodes(err);
    if ( fclean != found)
    {  err=ini_grid_cor();
       err=opt_grid_cor(fclean,err);
       }
    {  GRID_SUM *s;
       int j;
   	 for (j=0; j < grid_data_count; j++)
       {  s = grid_data + j;
          if (s->pix) free(s->pix);
          }
       free(grid_data);
       }
    printf( "figures are sent to xmgrace\n");	
    system ("xmgrace bm2grid.agr &");
   }
/*-----------------------------------------------------------*/
void read_grid_image (char * buffer)
{  char *c, *d; 
   int i, j, k;
   XY_HDR *hdr;
	IMAGE  *img;
	if (raw_img) free (raw_img);
   
   raw_img = img = (IMAGE*) prg_alloc(__FILE__, "raw->img", sizeof(IMAGE), 1);
	memcpy(img, &image, sizeof(IMAGE));
	
	grid_name = img->name = prg_get_name("@GRID_IMAGE", buffer);
   hdr = xy_open (img->name, prg_name, 0, r_def, 0, 1); 
   img->header = hdr;
   if (print) xy_puthdr( hdr, print & 1);
   xy_getsum(img, stdout, print, 1, 0 );
   xy_info(img, 0);
   xy_fputinfo( stdout, img);
	if (n_excluded)
	{  excl_image(img, print); 
      xy_info(img, 0);
      xy_fputinfo( stdout, img);
		}
	if (low_limit != -0x80000000) 
	for (j=0; j < img->y_size; j++)
   { for (i=0, k=j*img->x_size; i < img->x_size; i++, k++)  
	     if (img->data[k] < low_limit ) 
		  {  img->data[k] = low_limit;
		     img->frame[k] = 0;
	      }
	  }
	if (high_limit != -0x80000000) 
	for (j=0; j < img->y_size; j++)
   { for (i=0, k=j*img->x_size; i < img->x_size; i++, k++)  
	     if (img->data[k] > high_limit ) 
		  {  img->data[k] = high_limit;
		     img->frame[k] = 0;
	      }
	  }
	if ((low_limit != -0x80000000) || (high_limit != -0x80000000) )
   {  xy_info(img, 0);
      xy_fputinfo( stdout, img);
		}
   if (print & 256)
	{  printf("## map restricted to x,y<10\n");
      for (j=0; j < img->y_size && j < 10 ; j++)
      { for (i=0, k=j*img->x_size; i < img->x_size && i < 10; i++, k++)  
	       printf(" %4d %d", img->data[k], img->frame[k]);
	    printf("\n");
	  }  }
   }      
/*-----------------------------------------------------------*/
GRID_SUM * get_grid_data(void)
{  int alloc_sum = 0;
   int i,j,k,level;
   int count_sum=0;
   int tot_pix=0;
   int n_rej=0;
   GRID_PIXEL *pix;
   GRID_SUM *sum, *sum_n;
   double xo,yo;
   int inside;
   if ( grid_level ) level = grid_level;
   else abort();
   alloc_sum=40;
   sum = (GRID_SUM *) prg_alloc(__FILE__, "pix_sum", sizeof(GRID_SUM)*alloc_sum , 1);

	for (j=0; j < raw_img->y_size; j++)
   {  for (i=0, k=j*raw_img->x_size; i < raw_img->x_size; i++, k++)  
	   if (raw_img->frame[k] && raw_img->data[k] >level ) 
		{  sum_n = sum + count_sum;
         sum_n->count = 0;
         sum_n->alloc = 20;
         sum_n->pix = (GRID_PIXEL *) prg_alloc(__FILE__, "pix", sizeof(GRID_PIXEL)*sum_n->alloc, 1);
         inside =1;
         explore(sum_n, i, j, level, &tot_pix, &inside);
         if (inside && sum_n->count > grid_count) 
         {  count_sum++; 
            if (print & 16)
               printf( "%4d : sum_n->count=%d pixels found > level %d\n", count_sum, sum_n->count, level);
	         if (count_sum >= alloc_sum)
            {  alloc_sum+=40;
               sum = (GRID_SUM *) prg_realloc(__FILE__, "pix_sum", sum, sizeof(GRID_SUM)*alloc_sum );
               }
            }  
         else
         {  n_rej++;
            free (sum_n->pix);
            }
	      }
      }
   xo=world_x/2;
   yo=world_y/2;
   xy_freeimg(raw_img,0);
   printf( "%d pixels found > level %d, in %d used sums and %d rejected (count<=%d or located on borders)\n", 
       tot_pix, level, count_sum, n_rej, grid_count);
	for (j=0; j < count_sum; j++)
   {  sum_n = sum + j;
      pix = sum_n->pix;
      sum_n->x=0;
      sum_n->y=0;
      sum_n->z=0;
      for (i=0; i < sum_n->count; i++, pix++)
      {  sum_n->x+=pix->x*pix->z;
         sum_n->y+=pix->y*pix->z;
         sum_n->z+=pix->z;
         }
      sum_n->x=sum_n->x/sum_n->z;
      sum_n->y=sum_n->y/sum_n->z;
      sum_n->d=(sum_n->x-xo)*(sum_n->x-xo)+(sum_n->y-yo)*(sum_n->y-yo);
      sum_n->ok=1;
      if (print & 16)
         printf( "%4d : %10.3lf %10.3lf %9.1lf (%3d)  %8.2lf\n", 
              j, sum_n->x, sum_n->y, sum_n->z, sum_n->count, sum_n->d);
      }
   grid_data_count=count_sum;
   return sum;
   } 
/*-----------------------------------------------------------*/
int explore(GRID_SUM *sum,  int istart, int jstart, int level, int *tot_pix, int *inside)
{  int k;
   int n=1;
   GRID_PIXEL *pix;
   sum->count++;
   if (sum->count>=sum->alloc)
   {  sum->alloc+=10;
      sum->pix = (GRID_PIXEL *) prg_realloc(__FILE__, "pix_sum", sum->pix, sizeof(GRID_PIXEL)*sum->alloc );
      }
   pix=sum->pix+sum->count;
   if (istart==0 || istart == raw_img->x_size-1) *inside =0;
   if (jstart==0 || jstart == raw_img->y_size-1) *inside =0;
   pix->x=istart;
   pix->y=jstart;
   k=jstart*raw_img->x_size+istart;  
	pix->z=raw_img->data[k];
   raw_img->frame[k]=0;
   (*tot_pix)++;
   if (istart>=1)
   {  k=jstart*raw_img->x_size+istart-1;  
	   if (raw_img->frame[k] && raw_img->data[k] >level ) n+=explore(sum, istart-1, jstart, level, tot_pix, inside);
      }
   if (istart<raw_img->x_size-1)
   {  k=jstart*raw_img->x_size+istart+1;  
	   if (raw_img->frame[k] && raw_img->data[k] >level ) n+=explore(sum, istart+1, jstart, level, tot_pix, inside);
      }
   if (jstart<raw_img->y_size-1)
   {   k=(jstart+1)*raw_img->x_size+istart;  
	    if (raw_img->frame[k] && raw_img->data[k] >level ) n+=explore(sum, istart, jstart+1, level, tot_pix, inside);
       }
   return n;
   }
/*-----------------------------------------------------------*/
void find_grid_vectors ()
{  int i,j;
   GRID_SUM *s, *t;
   int start_j, trans_j;
   double mean=0;
   double d0=world_x*world_x+world_y*world_y;
   double start_x, start_y;
   double trans_x, trans_y;
   double ratio, rratio;
   double tx, ty, sx, sy, td, sd, sdo;
   double angle;
	for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      mean+=s->z;
      if (!s->ok) continue;
      if (s->d<d0 && j*s->z>mean/4)
      {  start_j=j;
         d0=s->d;
         }
      }
   mean/=grid_data_count;
   s = grid_data + start_j;
   start_x=s->x;
   start_y=s->y;
   s->nx=0;
   s->ny=0;
   printf( "grid %d spots, mean %.2lf, start %d-> %.2lf %.2lf (%.1lf %d)  d=%.1lf\n", 
       grid_data_count, mean, start_j, start_x, start_y, s->z, s->count, sqrt(s->d) );
   /* looking for translationa now */
	trans_j=0;
   trans_x=trans_y=0;
   sdo=world_x*world_x+world_y*world_y;
   for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      if (s->z < mean/4) continue;
      sd=sdo;
      for (i=0; i < grid_data_count; i++)
      {  if ( i == j) continue;
         t = grid_data + i;
         if (!t->ok) continue;
         if (t->z < mean/4) continue;
         if (t->x<=s->x) continue;
         tx=t->x-s->x;
         ty=t->y-s->y;
         if (ty>tx || ty+tx<=0) continue;
         td=tx*tx+ty*ty;
         if (td<sd)
         {  sx=tx;
            sy=ty;
            sd=td;
            }
          }
      if (sd != sdo)
      {   trans_j++;
          trans_x+=sx;
          trans_y+=sy;
          }
      }
   trans_x/=trans_j;
   trans_y/=trans_j;
   angle = 45/atan(1) * atan2(trans_y,trans_x);
   printf( "grid translation found %.2lf %.2lf  (mean of %d vector) angle %.2lf\n", 
         trans_x, trans_y, trans_j, angle );
   /* expecting a squared translation lattice */
   gridorig_x=start_x;      
   gridorig_y=start_y;      
   gridrot_xx=trans_x;      
   gridrot_xy=trans_y;      
   gridrot_yx=-gridrot_xy;      
   gridrot_yy=gridrot_xx; 
   gridnorm=sqrt(trans_x*trans_x+trans_y*trans_y);    
   }
/*-----------------------------------------------------------*/
int find_grid_nodes (double tol)
{  int j;
	double tx, ty;
   double x,y;
   double ex, ey, error;
   GRID_SUM *s;
   int found=0;
   for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      tx=s->x-gridorig_x;
      ty=s->y-gridorig_y;
      x=(tx*gridrot_xx+ty*gridrot_xy)/(gridnorm*gridnorm);
      y=(tx*gridrot_yx+ty*gridrot_yy)/(gridnorm*gridnorm);
      s->nx=(x>=0)?floor(x+0.5):floor(x+0.5);  
      s->ny=(y>=0)?floor(y+0.5):floor(y+0.5);  
      ex=x-s->nx;
      ey=y-s->ny;
      error=ex*ex+ey*ey;
      if (sqrt(error)<tol) { found++; s->ok=1; }
      else  s->ok=0;
      if (print & 8)
         printf( "grid coord %3d -> %3d %3d (%6.2lf %6.2lf) %s\n", 
               j,s->nx, s->ny, ex, ey, s->ok ? "valid":"rejected" );
      }
   printf( "grid coord %d spots found within %.2lf\n", 
               found, tol);
   return found;
   }
/*-----------------------------------------------------------*/
int clean_grid_nodes (double err)
{  int i, j, ir, ik;
   double sx, sy, es, tx, ty, et, er, ek, x, y;
   GRID_SUM *s, *t, *k, *r;
   int found, rej, loop=0;
   char *rmsg;
   do {
   rej=found=0;
   for (j=0; j < grid_data_count-1; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      for (i=j+1, t=s+1; i < grid_data_count; i++, t++)
      {  if (!t->ok) continue;
         if (s->nx == t->nx && s->ny == t->ny)
         {   found++; 
             x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
             y = gridorig_y + s->nx * gridrot_xy + s->ny * gridrot_yy;
             printf("nodes duplicated in %2d,%2d \n", s->nx,s->ny);
             sx = x - s->x;
             sy = y - s->y;
             es=sqrt(sx*sx+sy*sy);
             tx = x - t->x;
             ty = y - t->y;
             et=sqrt(tx*tx+ty*ty);
             if ( et > es) { r=t; k=s; ir=i; ik=j; er=et; ek=es; } 
             else          { k=t; r=s; ir=j; ik=i; er=es; ek=et; }
             if (er > 2*err) { rmsg="rejected"; rej++; r->ok=0; } else { rmsg="accepted"; } 
       /*      if (er < err ) */
             tx=r->x-gridorig_x;
             ty=r->y-gridorig_y;
             x=(tx*gridrot_xx+ty*gridrot_xy)/(gridnorm*gridnorm);
             y=(tx*gridrot_yx+ty*gridrot_yy)/(gridnorm*gridnorm);
             printf("  node %d (%d pixels,z=%5.0lf) %s, error=%.2lf  x=%.1lf(%.1lf)  y=%.lf(%.1lf) X=%.2lf Y=%.2lf\n",
                   ir, r->count, r->z, rmsg, er, r->x, r->cx, r->y, r->cy, x, y);
             tx=k->x-gridorig_x;
             ty=k->y-gridorig_y;
             x=(tx*gridrot_xx+ty*gridrot_xy)/(gridnorm*gridnorm);
             y=(tx*gridrot_yx+ty*gridrot_yy)/(gridnorm*gridnorm);
             printf("  node %d (%d pixels,z=%5.0lf) accepted, error=%.2lf  x=%.1lf(%.1lf)  y=%.lf(%.1lf) X=%.2lf Y=%.2lf\n",
                   ik, k->count, k->z, ek, k->x, k->cx, k->y, k->cy, x, y);
             }
          }
        }
      if (rej || !loop++) printf (" %d duplicates nodes found %d rejected\n", found, rej);
      } while (rej);
   found=0;
   for (j=0; j < grid_data_count-1; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      found++;
      }
  return found;
   }
/*-----------------------------------------------------------*/
double opt_grid_cor(int found, double err)
{  int i, k, sz, save_k;
   double errn, dfound = found;
   double save_xcor[COR], save_ycor[COR];
   double save_gridrot[4];
   double angle;
   save_k=k=2;
   sz=3;
   errn=err/sqrt((found-sz)/dfound);
   prt_grid_cor(save_xcor, grid_cor.xcor, "x", 3, 0, 0);
   prt_grid_cor(save_ycor, grid_cor.ycor, "y", 3, 0, 0);
   for (i=0;i<4;i++) save_gridrot[i]=gridrot[i];      
   do  
   {  err = errn;
      if ( ref_grid_spots(k,0) )
      {  double n=sqrt(grid_cor.xcor[1]*grid_cor.xcor[1]+grid_cor.xcor[2]*grid_cor.xcor[2]);
         for (i=0;i<4;i++) gridrot[i]/=n;      
         gridnorm=sqrt(gridrot_xx*gridrot_xx+gridrot_xy*gridrot_xy);
         printf(" Grid size adjust to %.2lf\n", gridnorm);
         angle = atan2(grid_cor.xcor[2],grid_cor.xcor[1]) + atan2(gridrot_xy,gridrot_xx);
         gridrot_xx=gridnorm*cos(angle);
         gridrot_yy=gridnorm*cos(angle);
         gridrot_xy=gridnorm*sin(angle);
         gridrot_yx=-gridnorm*sin(angle);
         }
      sz=ref_grid_spots(k,1);
      if (sz)
      {  cor_grid_spots(sz);
         errn=err_grid_spots(0)/sqrt((found-2*sz)/dfound);
        if (errn < err)
        {   save_k=k;
            prt_grid_cor(save_xcor, grid_cor.xcor, "x", sz, 0, 0);
            prt_grid_cor(save_ycor, grid_cor.ycor, "y", sz, 0, 0);
            for (i=0;i<4;i++) save_gridrot[i]=gridrot[i];      
            }
         }
      printf (" normalized err %.2f  %d nodes, size %d (%d)\n",errn , found, sz, k);
      k++;
      } while (errn < err && k <= DEG+1 );
   if ( k != save_k)
   {   double shift;
       k = save_k; 
       sz=(k*(k+1))/2;
       for (i=0;i<4;i++) gridrot[i]=save_gridrot[i];      
       gridnorm=sqrt(gridrot_xx*gridrot_xx+gridrot_xy*gridrot_xy);
       printf("\n Grid size set to %.2lf pixels (%.4lf)\n", gridnorm, grid_size/gridnorm);
       gridorig_x -= save_xcor[0] - xcor_orig;
       gridorig_y -= save_ycor[0] - ycor_orig;
       save_xcor[0] = xcor_orig;
       save_ycor[0] = ycor_orig;
       angle = 45/atan(1) * atan2(gridrot_xy,gridrot_xx);
       printf(" grid center adjust at (%.2lf,%.2lf), angle %.2lf\n", gridorig_x, gridorig_y, angle );
       prt_grid_cor(grid_cor.xcor, save_xcor, "x", sz, 0, 1);
       prt_grid_cor(grid_cor.ycor, save_ycor, "y", sz, 0, 1);
       cor_grid_spots(sz);
       errn=err_grid_spots(print&4?6:5)/sqrt((found-2*sz)/dfound); 
       printf (" normalized err %.2f  %d nodes, size %d\n",errn , found, sz);
       }
   grid_cor.size=sz;
   return errn;
   }
/*-----------------------------------------------------------*/
void cor_grid_spots(int size)
{  int j;
   double x,y;
   GRID_SUM *s;
   int found=0;
   s = grid_data;
   for (j=0; j < grid_data_count; j++, s++)
   {  x = s->x - xcor_orig;
      y = s->y - ycor_orig;
      s->cx = grid_cor_pos(x,y,grid_cor.xcor,size);
      s->cy = grid_cor_pos(x,y,grid_cor.ycor,size);
      }
   }
/*-----------------------------------------------------------*/
double err_grid_spots(int doprint)
{  int j;
	static double raw_error = 0;
   static double raw_sx =0; 
   static double raw_sy =0; 
   double ex, ey;
	double cerror = 0;
	double error_trig;
   double ecx, ecy;
   double scx =0; 
   double scy =0; 
   double x,y;
   GRID_SUM *s;
   int found=0;
   for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      found++; 
      x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
      y = gridorig_y + s->nx * gridrot_xy + s->ny * gridrot_yy;
      ex = x - s->x;
      ey = y - s->y;
      ecx = x - s->cx;
      ecy = y - s->cy;
      cerror+=ecx*ecx+ecy*ecy;
      scx+=ecx*ecx;
      scy+=ecy*ecy;
      if (doprint&2 || ( print & 16 && doprint&1) )
         printf( "grid coord %3d -> %3d %3d  : %6.1lf %6.1lf error %5.1lf %5.1lf\n", 
               j,s->nx, s->ny, s->x, s->y, ecx, ecy);
      }
   cerror=sqrt(cerror/found);
   if (raw_error == 0)
   {   raw_error = cerror;
       raw_sx = sqrt(scx/found);
       raw_sy = sqrt(scy/found);
       }
   printf( "grid coord %d spots mean error %.1lf pixels (raw %.1lf) x %.1lf(%.1lf) y %.1lf(%.1lf)\n", 
               found, cerror, raw_error, sqrt(scx/found), raw_sx, sqrt(scy/found), raw_sy );
   if (doprint  &4)
   {  error_trig=9*cerror*cerror;
      printf( "worst points (>3sigma):\n"); 
      for (j=0; j < grid_data_count; j++)
      {  s = grid_data + j;
        if (!s->ok) continue;
        found++; 
        x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
        y = gridorig_y + s->nx * gridrot_xy + s->ny * gridrot_yy;
        ex = x - s->x;
        ey = y - s->y;
        ecx = x - s->cx;
        ecy = y - s->cy;
        if ( ecx*ecx+ecy*ecy > error_trig )
        {  double tx=s->x-gridorig_x;
           double ty=s->y-gridorig_y;
           x=(tx*gridrot_xx+ty*gridrot_xy)/(gridnorm*gridnorm);
           y=(tx*gridrot_yx+ty*gridrot_yy)/(gridnorm*gridnorm);
           printf( "grid coord %3d -> (%2d,%2d) : X=%5.2lf Y=%5.2lf : x=%5.1lf erx=%4.1lf y=%5.1lf ery=%4.1lf\n", 
               j, s->nx, s->ny, x, y, s->x, ecx, s->y, ecy );
          }       
        }
     img_grid_spots(raw_img->name);   
     }
   return cerror;
   }
/*-----------------------------------------------------------*/
void img_grid_spots(char *name)
{  char fname[20];
   FILE *f;
   GRID_SUM *s;
   double x,y,e;
   int ix,iy,j;
   int minx =  world_x; 
   int maxx = -world_x;
   int miny =  world_y; 
   int maxy = -world_y;
   int zoomy=10;
   int ay;
   for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      if (s->nx > maxx) maxx = s->nx;
      if (s->ny > maxy) maxy = s->ny;
      if (s->nx < minx) minx = s->nx;
      if (s->ny < miny) miny = s->ny;
      }
   f=fopen ("bm2grid.agr","w");
   fprintf (f,"# Grace project file\n#\n@version 50104\n@page size 792, 612\n");
   fprintf (f,"@with string\n@ string on\n@ string loctype view\n@ string 0.2, 0.05\n");
   fprintf (f,"@ string def \"%s   --   %s\"\n", name, cvs_id);

   fprintf (f,"@g0 on\n@g0 hidden false\n");
   fprintf (f,"@with g0\n@ subtitle \"X distortions\"\n");
   fprintf (f,"@ world xmin 0\n@ world xmax 1260\n@ xaxis tick major 400\n");
   fprintf (f,"@ world ymin -100\n@ world ymax 100\n@ yaxis tick major 20\n");
   fprintf (f,"@ view xmin 0.09\n@ view xmax 0.59\n@ view ymin 0.15\n@ view ymax 0.85\n@ view ymax 0.85\n");
   fprintf (f,"@ legend 0.61, 0.75\n");
   for (iy=miny,ay=0;iy<=maxy;iy++,ay++)
   {  fprintf (f,"@ s%d symbol color %d\n@ s%d line color %d\n@ s%d line linewidth 3.0\n@ s%d legend \"%d\"\n",
        3*ay,1+ay,3*ay,1+ay,3*ay,3*ay,iy);
      fprintf (f,"@target G0.S%d\n@type xy\n",ay);
      for (ix=minx;ix<=maxx;ix++)
      {  for (j=0; j < grid_data_count; j++)
         {  s = grid_data + j;
            if (!s->ok) continue;
            if ( s->nx != ix ) continue;
            if ( s->ny != iy ) continue;
            x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
            e = s->cx - s->x;
            fprintf (f,"%8.0lf %8.2lf\n", x, e+zoomy*iy);
            }
         }
      fprintf (f,"&\n");
      fprintf (f,"@ s%d symbol 1\n@ s%d symbol color %d\n@ s%d line color %d\n",1+3*ay,1+3*ay,1+ay,1+3*ay,1+ay);
      fprintf (f,"@target G0.S%d\n@type xy\n",ay);
      for (ix=minx;ix<=maxx;ix++)
      {  for (j=0; j < grid_data_count; j++)
         {  s = grid_data + j;
            if (!s->ok) continue;
            if ( s->nx != ix ) continue;
            if ( s->ny != iy ) continue;
            x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
            e = x - s->x;
            fprintf (f,"%8.0lf %8.2lf\n", x, e+zoomy*iy);
            }
         }
      fprintf (f,"&\n");
      fprintf (f,"@ s%d line color %d\n",2+3*ay,1+ay);
      fprintf (f,"@target G0.S%d\n@type xy\n0 %d\n%d %d\n&\n",2+3*ay,zoomy*iy,world_x,zoomy*iy);
      }

   fprintf (f,"@g1 on\n@g1 hidden false\n");
   fprintf (f,"@with g1\n@ subtitle \"Y distortions\"\n");
   fprintf (f,"@ world xmin 0\n@ world xmax 1260\n@ xaxis tick major 400\n");
   fprintf (f,"@ world ymin -100\n@ world ymax 100\n@ yaxis tick major 20\n@ yaxis ticklabel place opposite\n");
   fprintf (f,"@ view xmin 0.73\n@ view xmax 1.23\n@ view ymin 0.15\n@ view ymax 0.85\n");
   for (iy=miny,ay=0;iy<=maxy;iy++,ay++)
   {  fprintf (f,"@ s%d symbol color %d\n@ s%d line color %d\n@ s%d line linewidth 3.0\n",
        3*ay,1+ay,3*ay,1+ay,3*ay);
      fprintf (f,"@target G1.S%d\n@type xy\n",3*ay);
      for (ix=minx;ix<=maxx;ix++)
      {  for (j=0; j < grid_data_count; j++)
         {  s = grid_data + j;
            if (!s->ok) continue;
            if ( s->nx != ix ) continue;
            if ( s->ny != iy ) continue;
            x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
            y = gridorig_y + s->nx * gridrot_xy + s->ny * gridrot_yy;
            e = s->cy - s->y;
            fprintf (f,"%8.0lf %8.2lf\n", x, e+zoomy*iy);
            }
         }
      fprintf (f,"&\n");
      fprintf (f,"@ s%d symbol 1\n@ s%d symbol color %d\n@ s%d line color %d\n",1+3*ay,1+3*ay,1+ay,1+3*ay,1+ay);
      fprintf (f,"@target G1.S%d\n@type xy\n",1+3*ay);
      for (ix=minx;ix<=maxx;ix++)
      {  for (j=0; j < grid_data_count; j++)
         {  s = grid_data + j;
            if (!s->ok) continue;
            if ( s->nx != ix ) continue;
            if ( s->ny != iy ) continue;
            x = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
            y = gridorig_y + s->nx * gridrot_xy + s->ny * gridrot_yy;
            e = y - s->y;
            fprintf (f,"%8.0lf %8.2lf\n", x, e+zoomy*iy);
            }
         }
      fprintf (f,"&\n");
      fprintf (f,"@ s%d line color %d\n",2+3*ay,1+ay);
      fprintf (f,"@target G1.S%d\n@type xy\n0 %d\n%d %d\n&\n",2+3*ay,zoomy*iy,world_x,zoomy*iy);
      }
   fclose(f);
   }
/*-----------------------------------------------------------*/
#define DCOR (2*COR)
int ref_grid_spots(int deg, int doprint)
{  int i,j, k;
	double error = 0;
   double ex, ey;
	double cerror = 0;
   double gx,gy,x,y;
   double matrice[DCOR*DCOR];
   double y_vect[DCOR];
   double der[DCOR];
   GRID_SUM *s;
   int cor = (deg*(deg+1))/2;
   if ( cor > COR )
   {   printf (" ref degree is too high! size %d deg %d\n", cor, deg);
       return 0;
       }
   for (i=0; i<DCOR;i++)  
   {  y_vect[i]=0;
      for (j=0; j <= i; j++) matrice[j+i*(i+1)/2]=0;
      }
   for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      gx = gridorig_x + s->nx * gridrot_xx + s->ny * gridrot_yx;
      x = s->x - xcor_orig;
      y = s->y - ycor_orig;
      for (i=0; i<cor;i++) der[i]=grid_cor_der(i,x,y,grid_cor.xcor);
      for (i=0; i<cor;i++) 
      {  y_vect[i]+=der[i]*gx;
         for (k=0; k <= i; k++) matrice[k+i*(i+1)/2]+=der[i]*der[k];
         }
      }
   for (j=0; j < grid_data_count; j++)
   {  s = grid_data + j;
      if (!s->ok) continue;
      gy = gridorig_y + s->nx * gridrot_xy + s->ny * gridrot_yy;
      x = s->x - xcor_orig;
      y = s->y - ycor_orig;
      for (i=0; i<cor;i++) der[i]=grid_cor_der(i,x,y,grid_cor.ycor);
      for (i=0; i<cor;i++) 
      {  y_vect[i+cor]+=der[i]*gy;
         for (k=0; k <= i; k++) matrice[k+cor+(i+cor)*(i+cor+1)/2]+=der[i]*der[k];
         }
      }
   if (chol_part (matrice, y_vect, 2*cor) == 0 ) 
     {    prt_grid_cor(grid_cor.xcor, y_vect, "x",cor, 0, doprint);
          prt_grid_cor(grid_cor.ycor, y_vect, "y",cor, cor, doprint);
          }
   return cor;
   }
/*-----------------------------------------------------------*/
#define MINFLOAT 1e-60
/* ---------------------------------------------------------------- */
int chol_part (double * mat, double * vect, int n)
    /* perform Choleski decomposition : M = A * D * tA
       with A(i,i=1) ,j>i=>A(i,j)=0 and D diagonal;
       as result  M-1 = tB * D-1 * B with B = A-1
       the input matrix is symetrical and must be defined
       the storage requirements are minimized
       the input matrix is given as its low triangular part A(i,j) i<=j
       cf L.Leon,F.Millet in TurboC mathematiques p190-193, Eyrolles(Paris) 1988
       */
{   register i, j, k;
    int ret_val = 0;
    double **rmat;
    double **amat;
    double *vtab, *dtab;
    /* this define minimize the storage requirements */
#define bmat  amat
#define new_b vtab
    /* array of ptr on the lines of the low triangular part of mat */
    rmat = (double **) prg_alloc (__FILE__, "inversion 0", n*sizeof(double *),0);
    for (i=0; i<n; i++) rmat[i] = mat + ((i * (i+1)) / 2);
    /* array of ptr's on allocated lines of a low triangular array */
    amat = (double **) prg_alloc (__FILE__, "inversion 1",n*sizeof(double *),0);
    for (i=0; i<n; i++) amat[i] = 
         (double *) prg_alloc (__FILE__, "inversion 2",(i+1)*sizeof(double),0);
    /* array of values */
    vtab = (double *) prg_alloc (__FILE__, "inversion 3",n*sizeof(double),0);
    dtab = (double *) prg_alloc (__FILE__, "inversion 4",n*sizeof(double),0);
    bmat[0][0] = 1;
    amat[0][0] = 1;
    dtab[0] = rmat[0][0];
    if (fabs (dtab[0]) > MINFLOAT )
    {   /* calculate D and A values */
	for (k=1; k<n; k++)
	{   for (i=0, dtab[k] = rmat[k][k]; i<k; i++)
	    {   for (j=0, vtab[i]=0; j<=i; j++)
		   vtab[i] += bmat[i][j] * rmat[k][j];
		amat [k][i] = vtab[i] / dtab[i];
		dtab[k] -= amat[k][i] * vtab[i];
		if (fabs(dtab[k]) < MINFLOAT) ret_val = k+1;
		}
	    /* calculate B values */
	    if (ret_val==0)
	    {   for (i=0; i<k; i++)
		    for (j=i, new_b[i]=0; j<k; j++)
		       new_b[i] -= bmat[j][i] * amat[k][j];
		for (i=0; i<k; i++) bmat[k][i] = new_b[i];
		amat [k][k] = 1;
	        }      
            }
	    /* calculate the lower part of the original matrix */
	    for (j=0; j<n; j++)
	     for (i=0; i<=j; i++)
	      for (k=j, rmat[j][i]=0; k<n; k++)
	       rmat[j][i] += bmat[k][i] * bmat[k][j] / dtab[k];
	    /* le second membre */
	    for (i=0;i<n;i++)
	    {  for (j=0,vtab[i]=0;j<i;j++) vtab[i] += rmat[i][j] * vect[j];
	       for (j=i;j<n;j++)           vtab[i] += rmat[j][i] * vect[j];
	       }
	    for (i=0;i<n;i++) vect[i] = vtab[i];
	    }
	else ret_val = 1;
	free(dtab);
	free(vtab);
	for (i=n-1; i>=0; i--) free (amat[i]);
	free(amat);
	free(rmat);
	return(ret_val);
	}
/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/
/*  routines used by "@GRID_PRINT" and "@GRID_INPUT"         */
/*-----------------------------------------------------------*/
void grid_print (char * buffer)
{  int i;
   char *fname = prg_get_name("@GRID_PRINT", buffer);
   FILE *f = fopen(fname,"w");
   fprintf(f,"#%s %s\n", prg_name, grid_name);
   fprintf(f,"GRID_DCOR=%d\n", grid_cor.size);
   fprintf(f,"GRID_XCOR=");
   for (i=0;i<grid_cor.size;i++) fprintf(f," %8lg",grid_cor.xcor[i]);
   fprintf(f,"\n");
   fprintf(f,"GRID_YCOR=");
   for (i=0;i<grid_cor.size;i++) fprintf(f," %8lg",grid_cor.ycor[i]);
   fprintf(f,"\n");
   fclose(f);   
   free(fname);
   }
/*-----------------------------------------------------------*/
void grid_input (char * b)
{  int i;
   char *fname = prg_get_name("@GRID_INPUT", b);
   FILE *f = fopen(fname,"r");
   char buffer[500], c , *d, *e;
   char *tag, *arg;
   grid_struct_ini(&grid_cor,COR);
   while (!feof(f))
   {   c = fgetc( f );
	    ungetc(c, f);
		 if (c  == '@') return;
	    fgets(buffer, sizeof(buffer)-1, f);
       if (feof(f)) break;
       d = buffer;
       while (*d == ' ' || *d == '\n' || *d == '\r') d++;
       if (*d=='#' || *d==0) continue;
       if ( prg_initag(d, &tag, &arg) == 0 ) continue;
       if ( grid_ini( &grid_cor, tag, arg)) continue;
       }
   fclose(f);   
   prt_grid_cor(grid_cor.xcor, grid_cor.xcor, "x", grid_cor.size, 0, 1 );
   prt_grid_cor(grid_cor.ycor, grid_cor.ycor, "x", grid_cor.size, 0, 1 );
   }
/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/
/*  routines used by "@GRID_CHECK"                           */
/*-----------------------------------------------------------*/
void grid_check (char * buffer)
{   int found, fclean;
    char b[200];
    double err;
    printf("GRID_CHECK : read_grid_image(%s)\n",buffer);
    read_grid_image (buffer);
    printf("GRID_CHECK : grid_cor_img\n");
    grid_cor_img ( &grid_cor, raw_img, 1 );
    strcpy(b,"grid_");
    strcat(b,raw_img->name);
    raw_img->name=b;
    edf_write(raw_img, t_mode);
    world_x=raw_img->x_size;
    xcor_orig=world_x/2;
    world_y=raw_img->y_size;
    ycor_orig=world_y/2;
    grid_data = get_grid_data();  
    find_grid_vectors();
    found = find_grid_nodes(grid_error);
    err=ini_grid_cor();
    err=opt_grid_cor(found,err);
    fclean = clean_grid_nodes(err);
    if ( fclean != found)
    {  err=ini_grid_cor();
       err=opt_grid_cor(fclean,err);
       }
    printf( "figures are sent to xmgrace\n");
    system ("xmgrace bm2grid.agr &");
   }
/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/
/* ---------------------------------------------------------------- */
   
